from setuptools import setup, find_packages

setup(
    name='PyMolecules',
    version='2.0.1',
    description='A Python package for all types of linked lists and most useful methods',
    long_description=open('README.txt').read()+'\n\n'+open('CHANGELOG.txt').read(),
    url='',
    author='Yokesh Bollineni',
    author_email='bollineniyokesh@gmail.com',
    license='MIT',
    keywords='linked lists',
    packages=find_packages(),
    install_requires=['']
)

